/* 
 *  This is a driver for the 
 *  NSM TrueSound Decoder (mpeg1 audio layer 3).
 *  This is version 1.2.1 of the driver.
 *
 *  Copyright (C) 1997 Magnus Berglund 
 *       (magnus@ior.campus.luth.se, bert@ludd.luth.se)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Version history:
 * 1.0.0  First released version
 * 1.1.0  Fixed a serious bug in the TRUESOUND_CONT ioctl. The bug could
 *        cause the system to hang. Thanks to Karsten Ballhorn for discovering 
 *	  this bug.
 * 1.2.0  The driver now supports up to three cards in one computer.
 * 1.2.1  Does now use one major number instead of three as in the 1.2.0 
 *        version. (There was _absolutly_ now reason to use three majors).
 *        Some other smaller changes. The sinout ioctl has been removed.
 *        The sinus generation didn't do anything usefull, so no reason to 
 *        to have it included (besides it didn't work very well). 
 * 1.2.2  _small_ changes to the tsd_play program. Karsten Ballhorn discovered 
 *        that the the driver doesn't work/compile with the early 2.0.x 
 *        kernels. (There was a change in the semaphores in the 2.0.24 kernel).
 *        
 * Thanks to:
 *   NSM AG for supplying me with information and hardware.
 *   Henrik Wallin, Fredrik Markstrm, Jakob Sandgren and Erik Lundgren for
 *    help and advise.
 *
 * ----
 * If you have any problems, questions or bug reports please don't 
 * hesitate to contact me at magnus@ior.campus.luth.se
 *
 */



/* === START CONFIG SECTION === */

#define TSD_MAJOR 60
/* ie. char-major-60 */

#define TSD_IO0 0x280
#define TSD_IO1 0x288
#define TSD_IO2 0x290
/* io must be on 8-byte-boundary, i.e. (io & ~0x7)==io */

#define TSD_IRQ0 10
#define TSD_IRQ1 11
#define TSD_IRQ2 12
/* possible irq values: 3, 5, 7, 10, 11, 12, 15 */

#define TSD_DMA0 5
#define TSD_DMA1 6
#define TSD_DMA2 7
/* possible dma values: 5, 6, 7 */

#define TSD_CARD_SEARCH_ADDR { 0x1FC, 0x1FD, 0x1FE, 0x1FF}
/* The addresses that the driver will lock for a card at initialization.
 * This is the address that are configurable on the card. 
 * You may set any of this values to zero if you don't want the driver 
 * to test that address (eg. if you have another sort of card at one of
 * the addresses)
 */

#define BUFFORDER 3
/* size of dma buffer, size=PAGESIZE*2^BUFFORDER
 * BUFFORDER=0 ->   4096 bytes
 * BUFFORDER=1 ->   8096 bytes
 * BUFFORDER=2 ->  16384 bytes
 * BUFFORDER=3 ->  32768 bytes
 * BUFFORDER=4 ->  65536 bytes
 * BUFFORDER=5 -> 131072 bytes (MAX)
 */

#undef BUFF_ALLOC_AT_OPEN
/* with BUFF_ALLOC_AT_OPEN defined the DMA buffers are allocated/freed on an 
 * open/release basis 
 */

#undef IRQ_DMA_ALLOC_AT_OPEN
/* define to allocate/dellocate IRQ and DMA on a open/release basis 
 */

/* === END CONFIG SECTION === */

#define LinuxVersionCode(v,p,s) (((v)<<16)+((p)<<8)+(s))

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/malloc.h>

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
#include <asm/uaccess.h>
#include <asm/poll.h>
#endif

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/dma.h>
#include <asm/page.h>
#include <asm/semaphore.h>

#include "ringbuff.h"
#include "truesound.h"
#include "layer2.h"
#include "volumetables.h"


static int io[3] = { TSD_IO0, TSD_IO1, TSD_IO2 };
static int irq[3] = { TSD_IRQ0, TSD_IRQ1, TSD_IRQ2 };
static int dma[3] = { TSD_DMA0, TSD_DMA1, TSD_DMA2 };
static int scan[4] = { 0, 0, 0, 0 };

static unsigned short tsd_card_addresses[] = TSD_CARD_SEARCH_ADDR;

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
MODULE_PARM(io, "1-3i");
MODULE_PARM(irq, "1-3i");
MODULE_PARM(dma, "1-3i");
MODULE_PARM(scan, "1-4i");
#endif

/* minor 'IOCTL_MINOR' can always do all the ioctl:s
 * The card also supports MPEG 1 layer 1/2 (after downloading some extra 
 * DSP-code to the card).
 */
#define IOCTL_MINOR 0 
#define MP1_MINOR 1
#define MP2_MINOR 2
#define MP3_MINOR 3
#define MAX_MINOR 3
#define ISPLAYERTYPE(devtype) ((devtype)!=IOCTL_MINOR)

#define LAYER1_2 0
#define LAYER3 2

#define BUFFSIZE (PAGE_SIZE * (1<<BUFFORDER))

#define DREG	0
#define CREG	6
#define CREG2	7

#define CARDNR(inode) ((MINOR((inode)->i_rdev)&0xC0) >> 6)
#define DEVTYPE(inode) (MINOR((inode)->i_rdev)&0x3F)

/* for debug messages */
#undef DEBUGIT
#define D_IOCTL     (1<<0)
#define D_WRACCESS  (1<<1)
#define D_OTHER     (1<<2)
#define D_SELECT    (1<<3)
#define DEBUGMASK D_WRACCESS|D_OTHER
#ifdef DEBUGIT
#define DEBUGK(level, args...) ({if ((level)&(DEBUGMASK)) { printk("%s: ", card->cardname); printk(args);} })
#else
#define DEBUGK(level, args...)
#endif

#define MIN(A, B) ((A)<(B)?(A):(B))
#define MAX(A, B) ((A)>(B)?(A):(B))

#define PRINTK(level,args...) ({ printk(level "%s: ",card->cardname); printk(args); })

#define LOCKLIST down(&card->list_mutex)
#define UNLOCKLIST up(&card->list_mutex)
#define LOCKOPEN down(&card->open_mutex)
#define UNLOCKOPEN up(&card->open_mutex)
#define LOCKI2C down(&card->i2c_mutex)
#define UNLOCKI2C up(&card->i2c_mutex)

typedef unsigned char uchar;


/* Every card has a list of 'writer' processes. 
 * A new process is placed first in the list. The first processes in the
 * list is the one that is allowed to play, all others are put to sleep.
 */
typedef struct write_queue_type {
#ifdef DEBUGIT
    int nr;
#endif
    int layer;
    int paused;
    struct wait_queue *wait_q;
    struct write_queue_type *next;
} write_queue_t;

#ifdef DEBUGIT
static int writer_count = 0;
#endif

static const uchar tsd_irq_bits[] = {
    0x0,			/* 0 (unused) */
    0x0,			/* 1 (unused) */
    0x0,			/* 2 (unused) */
    0x1 << 2,			/* 3 */
    0x0,			/* 4 (unused) */
    0x2 << 2,			/* 5 */
    0x0,			/* 6 (unused) */
    0x3 << 2,			/* 7 */
    0x0,			/* 8 (unused) */
    0x0,			/* 9 (unused) */
    0x4 << 2,			/* 10 */
    0x5 << 2,			/* 11 */
    0x6 << 2,			/* 12 */
    0x0,			/* 13 (unused) */
    0x0,			/* 14 (unused) */
    0x7 << 2			/* 15 */
};

static const uchar tsd_dma_bits[] = {
    0x0,				/* 0 (unused) */
    0x0,				/* 1 (unused) */
    0x0,				/* 2 (unused) */
    0x0,				/* 3 (unused) */
    0x0,				/* 4 (unused) */
    0x1,				/* 5 */
    0x2,				/* 6 */
    0x3,				/* 7 */
};

typedef struct {
    unsigned short io;
    unsigned int irq;
    unsigned int dma;
    unsigned short configuration_port; /* strapped card address */
    char *cardname;		/* used then allocating resources + messages */
    write_queue_t access_anchor; /* anchor for the list of writer processes */
    write_queue_t *current_access; /* the current writer process */
    int open_cnt;		/* to allocate/deallocate resources */
    int play_open_cnt; 
    char *dma_buffer;		/* DMA buffer */
    char *pending_dma_addr;	/* pointer to last started DMA-block */
    int pending_dma_len;	/* size of last started DMA-block */
    struct wait_queue *drain_wait_q;
    struct wait_queue *write_wait_q;
    int card_playing;
    int dma_initiated;
    uchar volume;
    uchar hardware_volume;
    uchar i2c_state;
    int current_layer;
    struct timer_list interrupt_timer;
    struct semaphore open_mutex;
    struct semaphore i2c_mutex;
    struct semaphore list_mutex;
    ringbuff_t rb; /* the ringbuffer */
} card_info;

static int nr_of_cards = 0;

#ifdef DEBUGIT
#define AINIT { 0, LAYER3, 0, NULL, NULL } 
#else
#define AINIT { LAYER3, 0, NULL, NULL } 
#endif
static card_info cards[] = {
    { 0, 0, 0, 0, "tsd0", AINIT, 
      &(cards[0].access_anchor), 
      0, 0, NULL, NULL, 0, NULL, NULL, 0, 0, 100, 0, 3, 
      LAYER3, {}, MUTEX, MUTEX, MUTEX, {} },
    
    { 0, 0, 0, 0, "tsd1", AINIT, 
      &(cards[1].access_anchor), 
      0, 0, NULL, NULL, 0, NULL, NULL, 0, 0, 100, 0, 3, 
      LAYER3, {}, MUTEX, MUTEX, MUTEX, {} },

    { 0, 0, 0, 0, "tsd2", AINIT, 
      &(cards[2].access_anchor), 
      0, 0, NULL, NULL, 0, NULL, NULL, 0, 0, 100, 0, 3, 
      LAYER3, {}, MUTEX, MUTEX, MUTEX, {} },
};

static card_info *irq_to_card[16];

/* starts the interrupt timeout timer */
extern inline void start_irq_timer(card_info *card);

/* handle dma */
extern inline void dma_stop(card_info *card);
static void dma_flush_buffers(card_info *card);	/* flush dma buffers */
extern inline void dma_set_up(card_info *card, char *addr, int size);
extern inline int dma_start_transfer(card_info *card);

/* set the board state */
extern inline void tsd_set_layer(card_info *card, int new_layer); 
static void tsd_stop(card_info *card);	/* stop playing */
static void tsd_start(card_info *card);	/* start playing */
static void tsd_reset_card(card_info *card);	/* reset the board */

static int get_writer_access(card_info *card, 
			     struct inode *inode, struct file *file,
			     write_queue_t *this);
extern inline write_queue_t *get_previous_node(card_info *card, 
					       write_queue_t *this);

extern inline void tsd_drain_buffers(card_info *card); 

static int tsd_load_l2_code(card_info *card);
static int tsd_set_volume(card_info *card, unsigned char att);

/* i2c functions */
#define wait20usecs udelay(20)
extern inline int i2c_data_in(card_info *card);
static int i2c_out(card_info *card, unsigned char data);
static void i2c_start(card_info *card);
static void i2c_stop(card_info *card);
static void i2c_clk_lo(card_info *card);
static void i2c_clk_hi(card_info *card);
static void i2c_data_lo(card_info *card);
static void i2c_data_hi(card_info *card);

static void tsd_interrupt_timeout(unsigned long arg);

/* file_operation functions */
static void tsd_interrupt(int irq, void *dev_id, struct pt_regs *regs);


#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)

static long tsd_write(struct inode * inode, struct file * file,
	      const char * buffer, unsigned long count);
static unsigned int tsd_poll(struct file *file,poll_table *wait);
static int tsd_release(struct inode *inode, struct file *file);

#else

static int tsd_write(struct inode * inode, struct file * file,
	      const char * buffer, int count);
static int tsd_select(struct inode *inode, struct file *file,
	       int sel_type, select_table *wait);
static void tsd_release(struct inode *inode, struct file *file);

#endif

static int tsd_ioctl(struct inode *inode, struct file *file, 
	      unsigned int cmd, unsigned long arg);
static int tsd_open(struct inode *inode, struct file *file);

/* allocate/deallocate irq and dma */
static int allocate_resources(card_info *card);
static void deallocate_resources(card_info *card);

/* helpers to init_module */
static int check_port_for_card(unsigned short adr, unsigned short io);
static int register_card(unsigned short adr, int cardnr);
static void unregister_card(card_info *card);
static int set_module_parameters(void);

void cleanup_module(void);


/* ------------ */

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static struct file_operations tsd_fops = {
    NULL,			/* tsd_seek */
    NULL,			/* tsd_read */
    tsd_write,			/* tsd_write */
    NULL,			/* tsd_readdir */
    tsd_poll,			/* tsd_poll */
    tsd_ioctl,			/* tsd_ioctl */
    NULL,			/* tsd_mmap */
    tsd_open,			/* tsd_open */
    tsd_release			/* tsd_release */
};
#else
static struct file_operations tsd_fops = {
    NULL,			/* tsd_seek */
    NULL,			/* tsd_read */
    tsd_write,			/* tsd_write */
    NULL,			/* tsd_readdir */
    tsd_select,			/* tsd_select */
    tsd_ioctl,			/* tsd_ioctl */
    NULL,			/* tsd_mmap */
    tsd_open,			/* tsd_open */
    tsd_release			/* tsd_release */
};
#endif




/*
 * Starts the irq timeout timer
 */
extern inline void start_irq_timer(card_info *card)
{
    init_timer(&card->interrupt_timer);
    card->interrupt_timer.function = tsd_interrupt_timeout;
    card->interrupt_timer.expires = jiffies + 5*HZ;
    card->interrupt_timer.data = (unsigned long)card;
    add_timer(&card->interrupt_timer);
}


/*
 * Stops the current DMA transfer
 */
extern inline void dma_stop(card_info *card)
{
    tsd_stop(card);
    if (card->dma_initiated) {
	disable_dma(card->dma);
	card->dma_initiated=0;
    }
}


/*
 * Stops the current dma transfer and clears the buffers
 */
static void dma_flush_buffers(card_info *card)
{
    unsigned long flags;
    DEBUGK(D_WRACCESS, "flushing buffers\n");
    save_flags(flags);
    cli();
    dma_stop(card);
    rb_init(&card->rb, card->dma_buffer, BUFFSIZE); 
    card->pending_dma_len = 0;
    restore_flags(flags);
    wake_up_interruptible(&card->write_wait_q);
    wake_up(&card->drain_wait_q);
}


/*
 * Sets up a DMA transfare of 'size' byte starting at address 'addr'
 * Call with interrupts disable.
 */
extern inline void dma_set_up(card_info *card, char *addr, int size)
{
    addr=(char*)virt_to_bus(addr);
    disable_dma(card->dma);
    clear_dma_ff(card->dma);
    set_dma_mode(card->dma, DMA_MODE_WRITE);
    set_dma_addr(card->dma, (int)addr);
    set_dma_count(card->dma, size);
    enable_dma(card->dma);
    card->dma_initiated=1;
}


/*
 * Sets up a new dma-transfer from ringbuffer
 * Call with interrupts disable.
 * returns 1 if a dma transfer was started
 */
extern inline int dma_start_transfer(card_info *card)
{
    card->pending_dma_len = BUFFSIZE/2;
    card->pending_dma_addr = rb_getblock(&card->rb, &card->pending_dma_len);
    card->pending_dma_len &= ~1; /* must transfer an even count of bytes */
    if (card->pending_dma_len>0) {
	dma_set_up(card, card->pending_dma_addr, card->pending_dma_len);
    }
    else {
	dma_stop(card);
	return 0;
    }
    return 1;
}


/*
 * Sets the layer on the card 
 * If layer2 is to be used then some code has to be downloaded to the card
 */
extern inline void tsd_set_layer(card_info *card, int new_layer) {
    if (LAYER3==card->current_layer && LAYER1_2==new_layer) {
	tsd_reset_card(card);
	tsd_load_l2_code(card);
    }
    else if (LAYER1_2==card->current_layer && LAYER3==new_layer) {
	tsd_reset_card(card); 
	tsd_set_volume(card, card->hardware_volume);
    }
    card->current_layer = new_layer;
}


/* 
 * Tells the card to stop playing
 */
static void tsd_stop(card_info *card)
{
    uchar creg;
    unsigned long flags;

    save_flags(flags);
    cli();
    do {
	if (!card->card_playing) break;
	creg = inb(card->io+CREG) & 0xE0;
	outb(creg, card->io+CREG);
	card->card_playing = 0;
	del_timer(&card->interrupt_timer);
    } while(0);
    restore_flags(flags);
}

/*
 * Tells the card to start playing
 */
static void tsd_start(card_info *card)
{
    uchar creg;
    uchar creg_play = tsd_dma_bits[card->dma] | tsd_irq_bits[card->irq];
    unsigned long flags;

    if (card->current_access->paused) return;
    tsd_set_layer(card, card->current_access->layer);
    save_flags(flags);
    cli();
    do {
	if (card->card_playing) break;
	if (!card->dma_initiated) {
	    if (0==dma_start_transfer(card)) break;
	}
	creg = inb(card->io+CREG) & 0xE0;
	outb(creg | creg_play, card->io+CREG);
	card->card_playing = 1;
	start_irq_timer(card);
    } while(0);
    restore_flags(flags);
}


/*
 * Does a on-board hardware reset
 */
static void tsd_reset_card(card_info *card)
{
    int i;

    tsd_stop(card);

    outb_p(0x80, card->io+CREG); /* reset */

    udelay(100);

    /* Wait until reset time is over (about 15ms) */
    for (i=0; i<1000; i++) {
	if (inb_p(card->configuration_port) != 0xFF) 
	    break;
	udelay(100);
	if (need_resched) schedule();
    }

    if (i==1000) {
	PRINTK(KERN_INFO, "reseting card...failed\n"); return;
    }

    /* configure io-base */
    outb((uchar) (card->io >> 3), card->configuration_port); 
    
    outb(0x3C, card->io+CREG);
    if (inb(card->io+CREG) != 0x3C) {
	PRINTK(KERN_INFO, "reseting card...failed\n"); return;
    }
    outb(0x18, card->io+CREG);
    if (inb(card->io+CREG) != 0x18) {
	PRINTK(KERN_INFO, "reseting card...failed\n"); return;
    }
    outb(0, card->io+CREG);
    if (inb(card->io+CREG) != 0x0) {
	PRINTK(KERN_INFO, "reseting card...failed\n"); return;
    }
    PRINTK(KERN_INFO, "reseting card...done\n");
    
    card->current_layer = LAYER3;

}


/*
 * Checks if 'this' is the current writer, if not: put the current process
 * to sleep until it has writer access
 */
static int get_writer_access(card_info *card, 
			     struct inode * inode, struct file *file,
			     write_queue_t *this) 
{
    LOCKLIST;
    if (this!=card->current_access) {
	DEBUGK(D_WRACCESS, "write_access #%d => SLEEP\n", this->nr);
	if (file->f_flags & O_NONBLOCK) return -EAGAIN;
	if (inode) up(&inode->i_sem);
	UNLOCKLIST;
	interruptible_sleep_on(&this->wait_q);
	if (current->signal & ~current->blocked) {
	    DEBUGK(D_WRACCESS, "write_access #%d SIGNALED\n", this->nr);
	    if (inode) down(&inode->i_sem);
	    return -EINTR;
	}
	if (inode) down(&inode->i_sem);
	DEBUGK(D_WRACCESS, "write_access #%d out of sleep\n", this->nr);
    }
    else
	UNLOCKLIST;
    return 0;
}


/*
 *  Returns the previous node in the writer access list
 */
extern inline write_queue_t *get_previous_node(card_info *card, 
					       write_queue_t *this)
{
    write_queue_t *prev;
    prev = card->current_access;
    while(prev!=NULL && prev->next!=this) {
	prev = prev->next;
    }
    if (prev==NULL) {
	PRINTK(KERN_ERR, "Ajajajaj... prev==NULL\n");
    }
    return prev;
}


/*
 * waits until the buffers are drained
 */
extern inline void tsd_drain_buffers(card_info *card)
{
    unsigned long flags;    
    save_flags(flags);
    cli();
    if (card->card_playing && card->dma_initiated) {
	restore_flags(flags);
	sleep_on(&card->drain_wait_q);
    }
    else {
	restore_flags(flags);
	dma_flush_buffers(card);
    }
}


/*
 * Downloads layer2 code to the card
 */
static int tsd_load_l2_code(card_info *card) {
    int cnt;

    LOCKI2C;
    
    PRINTK(KERN_INFO, "Loading layer 2 code...\n");
  
    i2c_start(card);
    i2c_out(card, 0x78);		/* address device on I2C Bus */
    i2c_out(card, 0x68);
    for (cnt=0; cnt<(sizeof(l2code)/sizeof(unsigned char)); cnt++) {
	if (i2c_out(card, l2code[cnt]) != 0)  {
	    i2c_stop(card); /* no ack */
	    PRINTK(KERN_INFO, "failed loading layer 2 code\n");
	    UNLOCKI2C;
	    return 1;
	}
	if (need_resched) schedule(); /* be nice */
    }
    i2c_stop(card);

    PRINTK(KERN_INFO, "layer 2 code loaded\n");

    UNLOCKI2C;

    return 0;
}


/*
 * Sets the volume, does not seam to work in layer2 mode
 */
static int tsd_set_volume(card_info *card, unsigned char att) {
    unsigned char cmd[] ={0x78,0x68,0x40,0x00};
    int count;

    LOCKI2C;
    cmd[3]=att;
    i2c_start(card);
    for (count=0; count<4; count++)
	if (i2c_out(card, cmd[count]) != 0) {
	    i2c_stop(card); /* oops, got no ack */
	    UNLOCKI2C;
	    return 1;
	}
    i2c_stop(card);
    UNLOCKI2C;
    return 0;
}


/*
 */
extern inline int i2c_data_in(card_info *card) {
    return( (unsigned char) (inb(card->io+CREG2)) & 0x04);
}


/*
 */
static int i2c_out(card_info *card, unsigned char data) {
    unsigned char mask;
    int i;
    i2c_clk_lo(card);
    for (i=0,mask=0x80; i<8; i++) {
	if (data & mask)
	    i2c_data_hi(card);
	else
	    i2c_data_lo(card);
	wait20usecs;
	i2c_clk_hi(card); 
	i2c_clk_lo(card);
	mask >>= 1;
    }
    i2c_data_hi(card);		/* don't drive line */
    wait20usecs;
    i2c_clk_hi(card);
    if (i2c_data_in(card)!=0)	{ /* Slave must pull down line */
	i2c_clk_lo(card);	/* ACK clock */
	return 1;		/* return error code */
    }
    i2c_clk_lo(card);		/* ACK clock */
    return 0;
}


/*
 */
static void i2c_start(card_info *card) {
    i2c_data_hi(card);
    i2c_clk_hi(card);
    i2c_data_lo(card);
    wait20usecs;
}


/*
 */
static void i2c_stop(card_info *card) {
    wait20usecs;
    i2c_data_lo(card);
    i2c_clk_hi(card);
    wait20usecs;
    i2c_data_hi(card);
}

/*
 */
static void i2c_clk_lo(card_info *card) {
    card->i2c_state &= 0xFE;
    outb(card->i2c_state, card->io+CREG2);
    wait20usecs;
}


/*
 */
static void i2c_clk_hi(card_info *card) {
    card->i2c_state |= 1;
    outb(card->i2c_state, card->io+CREG2);
    wait20usecs;
}


/*
 */
static void i2c_data_lo(card_info *card) {
    card->i2c_state &= 0xFD;
    outb(card->i2c_state, card->io+CREG2);
}


/*
 */
static void i2c_data_hi(card_info *card) {
    card->i2c_state |= 2;
    outb(card->i2c_state, card->io+CREG2);
}


/*
 * Called then an interupt timeout has been triggered 
 * This does must likely indicate some kind of hardware problem
 */
static void tsd_interrupt_timeout(unsigned long data)  
{
    card_info *card = (card_info *)data;
    if (card==NULL) {     /* paranoia test */
	printk(KERN_ERR "tsd: Aaaaaaaaaaaah... card==NULL");
	return;
    }
    PRINTK(KERN_ERR, "interrupt timeout!!\n");
    dma_flush_buffers(card);
    wake_up_interruptible(&card->write_wait_q);
    wake_up(&card->drain_wait_q);
}

/*
 * The interrupt routine marks the last written block as free 
 * then sets up a new dma transfer (if possible) 
 */
static void tsd_interrupt(int irq, void *dev_id, struct pt_regs *regs)  
{
    unsigned long flags;
    card_info *card = irq_to_card[irq];

    if (!card->card_playing) {
	PRINTK(KERN_ERR, "Unexpected interrupt\n");
    }

    save_flags(flags);
    cli();

    del_timer(&card->interrupt_timer);

    if (card->pending_dma_len>0)
	rb_freeblock(&card->rb, card->pending_dma_addr, card->pending_dma_len);

    (void)dma_start_transfer(card);

    restore_flags(flags);

    if (!card->dma_initiated) {
	wake_up(&card->drain_wait_q);
	DEBUGK(D_OTHER, "out of data\n");
    }
    else {
	start_irq_timer(card);
    }
    wake_up_interruptible(&card->write_wait_q);
}


/*
 * Write as much as possible to the ring buffer 
 * If the ring buffer is full then the process is put to sleep 
 * until there is space left in the buffer 
 */
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static long tsd_write(struct inode *inode, struct file *file,
	      const char *buffer, unsigned long count)
#else
static int tsd_write(struct inode *inode, struct file *file,
	      const char *buffer, int count)
#endif
{
    card_info *card = cards + CARDNR(inode);
    write_queue_t *waccess;
    unsigned int devtype = DEVTYPE(inode);
    unsigned long flags;
    char *dest_buff;
    int dest_size;
    int ret;

    if (!ISPLAYERTYPE(devtype)) return -EPERM;

    waccess = (write_queue_t *)file->private_data;
    ret = get_writer_access(card, inode, file, waccess);
    if (ret) return ret;

    ret = verify_area(VERIFY_READ, buffer, count);
    if (ret) return ret;
  
    if (rb_isfull(&card->rb)) {
	if (file->f_flags & O_NONBLOCK) return -EAGAIN;
	up(&inode->i_sem);
	interruptible_sleep_on(&card->write_wait_q);
	down(&inode->i_sem);
	if (current->signal & ~current->blocked) {
	    return -EINTR;
	}

	ret = get_writer_access(card, inode, file, waccess);
	if (ret) return ret;
    }

    dest_size = count;

    save_flags(flags);
    cli(); /* really needed ? */
    dest_buff = rb_putblock(&card->rb, &dest_size);
    restore_flags(flags);
    
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
    copy_from_user(dest_buff, buffer, dest_size);
#else
    memcpy_fromfs(dest_buff, buffer, dest_size);
#endif

    save_flags(flags);
    cli();
    rb_markblock(&card->rb, dest_size);
    restore_flags(flags);
    tsd_start(card);

    return dest_size;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)

/*
 * returns ready for write only if the current file * has access
 * and the ringbuffer is non-full
 */
static unsigned int tsd_poll(struct file *file, poll_table *wait)
{
    write_queue_t *waccess;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,48)
    card_info *card = cards + CARDNR(file->f_dentry->d_inode);
    unsigned int devtype = DEVTYPE(file->f_dentry->d_inode);
#else
    card_info *card = cards + CARDNR(file->f_inode);
    unsigned int devtype = DEVTYPE(file->f_inode);
#endif

    if (!ISPLAYERTYPE(devtype)) return -EPERM;

    LOCKLIST;
    waccess = (write_queue_t *)file->private_data;
    if (waccess && (waccess!=card->current_access)) {
	poll_wait(&waccess->wait_q, wait);
	UNLOCKLIST;
	return 0;
    }	

    poll_wait(&card->write_wait_q, wait);
    if (waccess!=card->current_access) { /* ioctl from non-writer */
	UNLOCKLIST;
	return 0;
    }
    UNLOCKLIST;
    if (rb_isfull(&card->rb))
	return 0;  /* Not ready yet */
    return POLLOUT | POLLWRNORM; /* ready for write */
}


#else


/*
 * returns ready for write only if the current file * has access
 * and the ringbuffer is non-full
 */
static int tsd_select(struct inode *inode, struct file *file,
	       int sel_type, select_table *wait)
{
    card_info *card = cards + CARDNR(inode);
    unsigned int devtype = DEVTYPE(inode);
    write_queue_t *waccess;

    if (!ISPLAYERTYPE(devtype)) return -EPERM;

    waccess = (write_queue_t *)file->private_data;
    DEBUGK(D_SELECT, "select #%d\n", waccess->nr);

    if (sel_type==SEL_OUT) {  /* ready for read? */

	LOCKLIST;
	if (card->current_access!=waccess) {
	    DEBUGK(D_SELECT, "select #%d not access => SLEEP\n",
		   waccess->nr);
	    UNLOCKLIST;
	    select_wait(&waccess->wait_q, wait);
	    DEBUGK(D_SELECT, "select #%d good morning\n",
		   waccess->nr);
	    return 0;
	}
	UNLOCKLIST;

	if (rb_isfull(&card->rb)) {
	    DEBUGK(D_SELECT, "select #%d buffer full => SLEEP\n",
		   waccess->nr);
	    select_wait(&card->write_wait_q, wait);
	    return 0;  /* Not ready yet */
	}

	return 1; /* ready */
    }
    else 
	return 0;  /* Never ready for read and exceptions */
}
#endif


/*
 * TRUESOUND_RESET AND TRUESOUND_SIN_TEST can only be done by the current
 * writer or on minor IOCTL_MINOR 
 */
static int tsd_ioctl(struct inode *inode, struct file *file, 
	      unsigned int cmd, unsigned long arg)
{
#define PLAYER_AND_NOT_CURRENT (ISPLAYERTYPE(devtype) && (waccess!=card->current_access))

    card_info *card = cards + CARDNR(inode);
    unsigned int devtype = DEVTYPE(inode);
    write_queue_t *waccess;
    uchar vol;
    int error, intval;
    int old_layer;

    waccess = (write_queue_t *)file->private_data;

    switch (cmd) {
    case TRUESOUND_RESET:
	DEBUGK(D_IOCTL, "ioctl(RESET)\n");
	LOCKLIST;
	if (PLAYER_AND_NOT_CURRENT) {
	    UNLOCKLIST;
	    return 0;
	}
	old_layer=card->current_layer;
	tsd_reset_card(card);
	tsd_start(card);
	UNLOCKLIST;
	break;
    case TRUESOUND_PAUSE: 
	DEBUGK(D_IOCTL, "ioctl(PAUSE)\n");
	LOCKLIST;
	if (PLAYER_AND_NOT_CURRENT) {
	    waccess->paused=1;
	    UNLOCKLIST;
	    break;
	}
	tsd_stop(card);
	card->current_access->paused = 1;
	UNLOCKLIST;
	break;
    case TRUESOUND_CONT: 	
	DEBUGK(D_IOCTL, "ioctl(CONT)\n");
	LOCKLIST;
	if (PLAYER_AND_NOT_CURRENT) {
	    waccess->paused=0;
	    UNLOCKLIST;
	    break;
	}
	card->current_access->paused = 0;
	tsd_start(card);
	UNLOCKLIST;
	break;
    case TRUESOUND_SETVOL: 
	error = verify_area(VERIFY_READ, (void *)arg, 
			    sizeof(uchar));
	if (error) return error;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	get_user(vol,(uchar *)arg);	
#else
	vol = get_user((uchar *)arg);
#endif	
	vol = MIN(vol, 100);
	card->hardware_volume = volume_to_hardware[vol];
	card->volume = vol;
	tsd_set_volume(card, card->hardware_volume);
	DEBUGK(D_IOCTL, "ioctl(SETVOL, %d)\n", vol);
	break;
    case TRUESOUND_GETVOL: 
	DEBUGK(D_IOCTL, "ioctl(GETVOL)\n");
	error = verify_area(VERIFY_WRITE, (void *)arg, 
			    sizeof(uchar));
	if (error) return error;
	vol = card->volume;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	return put_user((uchar)vol, (uchar *)arg);
#else
	put_user((uchar)vol, (uchar *)arg);
	return 0;
#endif

    case TRUESOUND_GETSYNC:
	error = verify_area(VERIFY_WRITE, (void *)arg, 
			    sizeof(int));
	if (error) return error;
	intval = (inb(card->io+CREG2) & 0x8) >> 3;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	return put_user((int)intval, (int *)arg);
#else
	put_user((int)intval, (int *)arg);
	return 0;
#endif

    case TRUESOUND_GETCRC:
	error = verify_area(VERIFY_WRITE, (void *)arg, 
			    sizeof(int));
	if (error) return error;
	intval = ! ( (inb(card->io+CREG2) & 0x10) >> 4 );
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	return put_user((int)intval, (int *)arg);
#else
	put_user((int)intval, (int *)arg);
	return 0;
#endif

    case TRUESOUND_FLUSH_BUFFERS:
	LOCKLIST;
	if (PLAYER_AND_NOT_CURRENT) {
	    UNLOCKLIST;
	    break;
	}
	DEBUGK(D_IOCTL, "ioctl(TRUESOUND_FLUSH_BUFFERS)\n");
	dma_flush_buffers(card);
	UNLOCKLIST;
	break;

    case TRUESOUND_GET_BUFFSIZE:
	DEBUGK(D_IOCTL, "ioctl(GETVOL_BUFFSIZE)\n");
	error = verify_area(VERIFY_WRITE, (void *)arg, 
			    sizeof(unsigned long));
	if (error) return error;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	return put_user((unsigned long)BUFFSIZE, (unsigned long *)arg);
#else
	put_user((unsigned long)BUFFSIZE, (unsigned long *)arg);
	return 0;
#endif

    case TRUESOUND_GET_PAUSED:
	DEBUGK(D_IOCTL, "ioctl(GETVOL_BUFFSIZE)\n");
	error = verify_area(VERIFY_WRITE, (void *)arg, 
			    sizeof(int));
	if (error) return error;
	LOCKLIST;
	if (PLAYER_AND_NOT_CURRENT) {
	    intval = waccess->paused;
	    UNLOCKLIST;
	}
	else {
	    UNLOCKLIST;
	    intval = card->current_access->paused;
	}
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	return put_user(intval, (int *)arg);
#else
	put_user(intval, (int *)arg);
	return 0;
#endif
    
    default:
	return -EINVAL;
    }
    return 0;
}


/*
 * open on minor IOCTL_MINOR is simple
 * The other minors must be added to the writer list
 * and the correct layer must be set
 * file->private_data sets to point to the node in the list
 */
static int tsd_open(struct inode *inode, struct file *file)
{
    card_info *card;
    unsigned int devtype = DEVTYPE(inode);
    int error;

    if (CARDNR(inode)>=nr_of_cards) return -ENXIO;
    if (devtype>MAX_MINOR) return -ENXIO;
    card = cards + CARDNR(inode);

    LOCKOPEN;

    if (card->open_cnt==0) {
	error = allocate_resources(card);
	if (error) {
	    UNLOCKOPEN;
	    return error;
	}
    }

    if (ISPLAYERTYPE(devtype)) {
	write_queue_t *new_access_node;

	new_access_node = (write_queue_t *)kmalloc(sizeof(write_queue_t),
						    GFP_KERNEL);
	if (new_access_node == NULL) {
	    deallocate_resources(card);
	    UNLOCKOPEN;
	    return -ENOMEM;
	}
#ifdef DEBUGIT
	new_access_node->nr = writer_count++;
#endif
	new_access_node->wait_q = NULL;
	new_access_node->paused = 0;

	switch(devtype) {
	case MP1_MINOR : new_access_node->layer = LAYER1_2; break;
	case MP2_MINOR : new_access_node->layer = LAYER1_2; break;
	case MP3_MINOR : new_access_node->layer = LAYER3; break;
	default : new_access_node->layer = LAYER3; break;
	}

	wake_up(&card->drain_wait_q);

	LOCKLIST;
	new_access_node->next = card->current_access;
	card->current_access = new_access_node;
 	file->private_data = (void *)new_access_node;
	UNLOCKLIST;

	dma_flush_buffers(card); 

	DEBUGK(D_WRACCESS, "new writer #%d\n", new_access_node->nr);
	card->play_open_cnt++;

    }
    else {
	file->private_data = NULL;
    }

    card->open_cnt++;
    MOD_INC_USE_COUNT;

    UNLOCKOPEN;

    return 0;
}


/*
 * If it is the current writer that does the release
 * then first drain the buffers then activate the next writer.
 * If it not the first then just unlink the node from the list
 */
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static int tsd_release(struct inode *inode, struct file *file)
#else
static void tsd_release(struct inode *inode, struct file *file)
#endif
{
    card_info *card = cards + CARDNR(inode);
    unsigned int devtype = DEVTYPE(inode);
    write_queue_t *this_node;
    write_queue_t *prev_node;

    if (ISPLAYERTYPE(devtype)) { 
	this_node = (write_queue_t *)file->private_data;
	DEBUGK(D_WRACCESS, "release #%d\n", this_node->nr);
	
	LOCKLIST;
	if (this_node!=card->current_access) {
	    prev_node = get_previous_node(card, this_node);
	    if (prev_node!=NULL)
		prev_node->next = this_node->next;
	}
	else {
	    if (file->f_flags & O_NONBLOCK)
		dma_flush_buffers(card);
	    else
		tsd_drain_buffers(card);
	    card->current_access = card->current_access->next;
	    wake_up_interruptible( &(card->current_access->wait_q) );
	}
	kfree(this_node);
	UNLOCKLIST;
	card->play_open_cnt--;
    }
    LOCKOPEN;
    MOD_DEC_USE_COUNT;
    card->open_cnt--;
    deallocate_resources(card);
    UNLOCKOPEN;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
    return 0;
#endif
}


/*
 * Allocates irq and dma
 */
static int allocate_resources(card_info *card) {

#ifdef IRQ_DMA_ALLOC_AT_OPEN
    int error;
    error = request_irq(card->irq, tsd_interrupt, SA_INTERRUPT, 
			card->cardname, NULL);
    if (error) {
	PRINTK(KERN_ERR, "couldn't allocate irq %d\n", card->irq);
	return -EBUSY;
    }
    error = request_dma(card->dma, card->cardname);
    if (error) {
	free_irq(card->irq, NULL);
	PRINTK(KERN_ERR, "couldn't allocate dma %d\n", card->dma);
	return -EBUSY;
    }
#endif
    
#ifdef BUFF_ALLOC_AT_OPEN	
    card->dma_buffer = (uchar*)__get_dma_pages(GFP_DMA,BUFFORDER);
    if (card->dma_buffer == NULL) {
	free_dma(card->dma);
	free_irq(card->irq, NULL);
	PRINTK(KERN_ERR, "couldn't allocate DMA buffers\n");
	return -ENOMEM;
    }
#endif

    return 0;
}


/*
 *  dealocates irq and dma
 */
static void deallocate_resources(card_info *card) {
#ifdef IRQ_DMA_ALLOC_AT_OPEN
    if (card->open_cnt==0) {
	free_dma(card->dma);
	free_irq(card->irq, NULL);
    }
#endif
#ifdef BUFF_ALLOC_AT_OPEN
    if (card->open_cnt==0) {
	free_pages((unsigned int)card->dma_buffer,BUFFORDER);
    }
#endif
}


/*
 * checks the port 'adr' for a TrueSound decode card
 * io is the io-addr the card should be initilalized to.
 * returns 1 if a card is found, 0 otherwise.
 */
static int check_port_for_card(unsigned short adr, unsigned short io) {
    if (check_region(adr,1)) { /* busy */
	printk(KERN_ERR "(address 0x%x is busy)\n", adr); 
	return 0;
    }
    if (inb(adr) != 0xFF) { 
	/* Try to initiate L3 controller on io */
	outb((uchar) (io >> 3), adr);
	
	/* OK, now do some access testing */
	outb(0x3C, io+CREG);
	if (inb(io+CREG) != 0x3C)
	    return 0;
	outb(0x18, io+CREG);
	if (inb(io+CREG) != 0x18)
	    return 0;
	outb(0, io+CREG);
	if (inb(io+CREG) != 0x0)
	    return 0;
	return 1;

    }
    return 0;
}


/*
 * registers the io addresses for the card on io-address 'adr',
 * registers the irq/dma and allocates dma memory (if the xx_ALLOC_AT_OPEN
 * defines are not set)
 */
static int register_card(unsigned short adr, int cardnr) {
    card_info *card = cards + cardnr;

#ifndef IRQ_DMA_ALLOC_AT_OPEN
    int error;
    error = request_irq(card->irq, tsd_interrupt, SA_INTERRUPT, 
			card->cardname, NULL);
    if (error) {
	PRINTK(KERN_ERR, "could not allocate irq %d\n", card->irq);
	return -EBUSY;
    }
    error = request_dma(card->dma, card->cardname);
    if (error) {
	free_irq(card->irq, NULL);
	PRINTK(KERN_ERR, "could not allocate dma %d\n", card->dma);
	return -EBUSY;
    }
#endif
#ifndef BUFF_ALLOC_AT_OPEN
    card->dma_buffer = (uchar*)__get_dma_pages(GFP_DMA,BUFFORDER);
    if (card->dma_buffer == NULL) {
#ifndef IRQ_DMA_ALLOC_AT_OPEN
	free_dma(card->dma);
	free_irq(card->irq, NULL);
#endif
	printk(KERN_ERR "tsd: couldn't allocate DMA buffers\n");
	return -ENOMEM;
    }
#endif

    card->configuration_port = adr;
    request_region(card->io, 8, card->cardname);
    request_region(card->configuration_port, 1, card->cardname); 
    PRINTK(KERN_INFO, "card found at 0x%x, using: irq=%d,dma=%d,io=0x%x\n", 
	   card->configuration_port, card->irq, card->dma, card->io);

    return 0;
}


/*
 * Undo's what register_card does.
 */
static void unregister_card(card_info *card) {
#ifndef IRQ_DMA_ALLOC_AT_OPEN
    free_dma(card->dma);
    free_irq(card->irq, NULL);
#endif
#ifndef BUFF_ALLOC_AT_OPEN
    free_pages((unsigned int)card->dma_buffer,BUFFORDER);
#endif
    release_region(card->io, 8);
    release_region(card->configuration_port, 1);
}


/* 
 * Checks the module parameters and sets the values in the 'cards' struct 
 */
static int set_module_parameters(void) {
    int i;
    card_info *card;

    if (scan[0]!=0) {
	for (i=0; i<4; i++) {
	    tsd_card_addresses[i] = (unsigned short)scan[i];
	}
    }	
    for (i=0; i<4; i++) {
	if (tsd_card_addresses[i]!=0 
	    && (tsd_card_addresses[i]<0x1fc || tsd_card_addresses[i]>0x1ff)) {
	    printk(KERN_ERR "tsd: 0x%x is an invalid card address\n", 
		   tsd_card_addresses[i]);
	    printk(KERN_ERR "tsd: valid addresses are: "
		   "0x1fc, 0x1fd, 0x1fe, and 0x1ff\n");
	    return -EINVAL;
	}
    }

    for (i=0; i<3; i++) {
	card = &cards[i];
	card->io = (unsigned short)io[i];
	card->irq = (unsigned int)irq[i];
	card->dma = (unsigned int)dma[i];
	if ( (card->io & ~0x7)!=card->io) {
	    PRINTK(KERN_ERR, "invalid io address (%d)\n"
		   "io must start on 8-byte-boundary\n"
		   "    i.e. (io & ~0x7)==io)\n", card->io);
	    return -EINVAL;
	}
	
	if ( (card->irq<0||card->irq>15) || (tsd_irq_bits[card->irq]==0) ) {
	    PRINTK(KERN_ERR, "invalid irq (%d)\n"
		   "    possible values are: 3, 5, 7, 10, 11, 12, 15\n",
		   card->irq);
	    return -EINVAL;
	}
	
	if ( (card->dma<0||card->dma>7) || (tsd_dma_bits[card->dma]==0) ) {
	    PRINTK(KERN_ERR, "invalid dma (%d)\n"
		   "    possible values are: 5, 6, 7\n", card->dma);
	    return -EINVAL;
	}
    }
    return 0;
}


/*
 */
int init_module(void)
{
    int i, ret;
    int card_found=0;

#ifdef DEBUGIT 
    printk(KERN_INFO "tsd driver compiledate %s, %s\n", __DATE__, __TIME__);
#endif

    ret = set_module_parameters();
    if (ret) return ret;

    ret=register_chrdev(TSD_MAJOR, 
			"tsd",
			&tsd_fops);
    if (ret) {
	printk(KERN_ERR "tsd error: couldn't register char-major-%d!\n", 
	       TSD_MAJOR);
	return -EBUSY;
    }
    
    printk(KERN_INFO "TrueSoundDecoder: searching card(s):\n");
    for (i=0; i<4; i++) {
	/* try all "configuartion" ports 
	 * if we can read other than FF, then try to config
	 * the cards baseaddr */
	if (tsd_card_addresses[i]==0) continue;
	if (check_region(cards[nr_of_cards].io, 8)) {
	    printk(KERN_ERR "tsd error: ioports 0x%x->0x%x already in use!!\n",
		   cards[nr_of_cards].io, cards[nr_of_cards].io+7);
	    continue;
	}
	if (check_port_for_card(tsd_card_addresses[i],
				cards[nr_of_cards].io)) {
	    ret=register_card(tsd_card_addresses[i], nr_of_cards);
	    if (ret) {
		cleanup_module();
		return ret;
	    }
	    card_found=1;
	    nr_of_cards++;
	    if (nr_of_cards==3) break;/* for those who have four cards... */
	}
    }
    
    if (!card_found) {
	printk(KERN_INFO "tsd: no TrueSoundDecoder card found\n");
	cleanup_module();
	return -ENODEV;
    }

    /* initialize the irq_to_card table */
    for (i=0; i<16; i++) 
	irq_to_card[i] = NULL;
    for (i=0; i<nr_of_cards; i++)
	irq_to_card[ cards[i].irq ] = cards+i;

    /* reset the card(s) */
    for (i=0; i<nr_of_cards; i++) {
	tsd_reset_card(cards+i);
	tsd_set_volume(cards+i, cards[i].hardware_volume);
    }


    
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
    EXPORT_NO_SYMBOLS;
#else
    register_symtab(0); /* do not export any symbols */
#endif

    return 0;
}


/*
 */
void cleanup_module(void) 
{
    int i;
    printk(KERN_INFO "cleaning up tsd module\n");
    for (i=0; i<nr_of_cards; i++) {
	unregister_card(&cards[i]);
    }
    unregister_chrdev(TSD_MAJOR, "tsd");
}
    
